# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
"""Base components of HNCcorr."""


from copy import deepcopy

from hnccorr.movie import Patch
from hnccorr.graph import (
    CorrelationEmbedding,
    exponential_distance_decay,
    GraphConstructor,
    SparseComputationEmbeddingWrapper,
)
from hnccorr.seeds import (
    PositiveSeedSelector,
    NegativeSeedSelector,
    LocalCorrelationSeeder,
)
from hnccorr.segmentation import HncParametricWrapper
from hnccorr.postprocessor import SizePostprocessor


class Candidate:
    """Encapsulates the logic for segmenting a single cell candidate / seed.

    Attributes:
        best_segmentation (Segmentation): Segmentation of a cell's spatial footprint as
            selected by the postprocessor.
        center_seed (tuple): Seed pixel coordinates.
        clean_segmentations (list[Segmentation]): List of segmentation after calling
            `clean()` on each segmentation.
        segmentations (list[Segmentation]): List of segmentations returned by HNC.
        _hnccorr (HNCcorr): HNCcorr object.
    """

    def __init__(self, center_seed, hnccorr):
        """Initialize Candidate object."""
        self.center_seed = center_seed
        self._hnccorr = hnccorr
        self.segmentations = None
        self.clean_segmentations = None
        self.best_segmentation = None

    def __eq__(self, other):
        """Compare Candidate object."""
        # pylint: disable=W0212
        if isinstance(other, Candidate):
            return (self.center_seed == other.center_seed) and (
                self._hnccorr == other._hnccorr
            )

        return False

    def segment(self):
        """Segment candidate cell and return footprint (if any).

        Encapsulates the procedure for segmenting a single cell candidate. It
        determines the seeds, constructs the similarity graph, and solves the HNC
        clustering problem for all values of the trade-off parameter lambda. The
        postprocessor selects the best segmentation or determines that no cell is found.

        Returns:
            Segmentation or None: Best segmentation or None if no cell is found.
        """
        movie = self._hnccorr.movie
        pos_seeds = self._hnccorr.positive_seed_selector.select(self.center_seed, movie)
        neg_seeds = self._hnccorr.negative_seed_selector.select(self.center_seed, movie)
        patch = self._hnccorr.patch_class(
            movie, self.center_seed, self._hnccorr.patch_size
        )
        embedding = self._hnccorr.embedding_class(patch)
        graph = self._hnccorr.graph_constructor.construct(patch, embedding)
        self.segmentations = self._hnccorr.segmentor.solve(graph, pos_seeds, neg_seeds)
        self.clean_segmentations = [
            s.clean(pos_seeds, movie.pixel_shape) for s in self.segmentations
        ]
        self.best_segmentation = self._hnccorr.postprocessor.select(
            self.clean_segmentations
        )
        return self.best_segmentation


class HNCcorr:
    """Implementation of the HNCcorr algorithm.

    This class specifies all components of the algoritm and defines the procedure for
    segmenting the movie. How each candidate seed / location is evaluated is specified
    in the Candidate class.

    References:
        Q Spaen, R Asín-Achá, SN Chettih, M Minderer, C Harvey, and DS Hochbaum (2019).
        HNCcorr: A Novel Combinatorial Approach for Cell Identification in
        Calcium-Imaging Movies. eNeuro, 6(2).

    """

    def __init__(
        # pylint: disable=C0330
        self,
        seeder,
        postprocessor,
        segmentor,
        positive_seed_selector,
        negative_seed_selector,
        graph_constructor,
        candidate_class,
        patch_class,
        embedding_class,
        patch_size,
    ):
        """Initalizes HNCcorr object."""
        self.seeder = seeder
        self.postprocessor = postprocessor
        self.segmentor = segmentor
        self.positive_seed_selector = positive_seed_selector
        self.negative_seed_selector = negative_seed_selector
        self.graph_constructor = graph_constructor
        self._candidate_class = candidate_class
        self.patch_class = patch_class
        self.embedding_class = embedding_class
        self.patch_size = patch_size

        self.movie = None
        self.segmentations = []
        self.candidates = []

    @classmethod
    def from_config(cls, config=None):
        """Initializes HNCcorr from an HNCcorrConfig object.

        Provides a simple way to initialize an HNCcorr object from a configuration.
        Default components are used, and parameters are taken from the input
        configuration or inferred from the default configuration if not specified.

        Args:
            config (HNCcorrConfig): HNCcorrConfig object with modified configuration.
                Parameters that are not explicitly specified in the `config` object are
                taken from the default configuration ``DEFAULT_CONFIGURATION`` as
                defined in the `hnccorr.config` module.

        Returns:
            HNCcorr: Initialized HNCcorr object as parametrized by the configuration.
        """
        if config is None:
            config = DEFAULT_CONFIG
        else:
            config = DEFAULT_CONFIG + config

        edge_selector = SparseComputationEmbeddingWrapper(
            config.sparse_computation_dimension, config.sparse_computation_grid_distance
        )

        return cls(
            LocalCorrelationSeeder(
                config.seeder_mask_size,
                config.percentage_of_seeds,
                config.seeder_exclusion_padding,
                config.seeder_grid_size,
            ),
            SizePostprocessor(
                config.postprocessor_min_cell_size,
                config.postprocessor_max_cell_size,
                config.postprocessor_preferred_cell_size,
            ),
            HncParametricWrapper(0, 1),
            PositiveSeedSelector(config.positive_seed_radius),
            NegativeSeedSelector(
                config.negative_seed_circle_radius, config.negative_seed_circle_count
            ),
            GraphConstructor(
                edge_selector,
                lambda a, b: exponential_distance_decay(
                    a, b, config.gaussian_similarity_alpha
                ),
            ),
            Candidate,
            Patch,
            CorrelationEmbedding,
            config.patch_size,
        )

    def segment(self, movie):
        """Applies the HNCcorr algorithm to identify cells in a calcium-imaging movie.

        Identifies cells the spatial footprints of cells in a calcium imaging movie.
        Cells are identified based on a set of candidate locations identified by the
        seeder. If a cell is found, the pixels in the spatial footprint are excluded as
        seeds for future segmentations. This prevents that a cell is segmented more
        than once. Although segmented pixels cannot seed a new segmentation, they may
        be segmented again.

        Identified cells are accessible through the `segmentations` attribute.

        Returns:
            Reference to itself.
        """
        self.movie = movie
        self.seeder.reset()
        self.segmentations = []
        self.candidates = []

        self.seeder.select_seeds(movie)

        seed = self.seeder.next()
        while seed is not None:
            candidate = self._candidate_class(seed, self)
            self.candidates.append(candidate)
            print(
                "Cells identified: %d, Next candidate: %d"
                % (len(self.segmentations), len(self.candidates))
            )
            best_segmentation = candidate.segment()
            if best_segmentation is not None:
                self.segmentations.append(best_segmentation)
                self.seeder.exclude_pixels(best_segmentation.selection)
            seed = self.seeder.next()

        print("Completed - Total cells identified: %d" % len(self.segmentations))
        return self

    def segmentations_to_list(self):
        """Exports segmentations to a list of dictionaries.

        Each dictionary in the list corresponds to the footprint of a cell. Each
        dictionary contains the key `coordinates` containing a list of pixel
        coordinates. Each pixel coordinate is a tuple with the zero-indexed coordinates
        of the pixel. Pixels are indexed like matrix coordinates.

        Returns
            list[dict[tuple]]: List of cell coordinates.

        """
        output = []
        for segmentation in self.segmentations:
            output.append({"coordinates": list(segmentation.selection)})
        return output


class HNCcorrConfig:
    """Configuration class for HNCcorr algorithm.

    Enables tweaking the parameters of HNCcorr when used with the default components.
    Configurations are modular and can be combined using the addition operation.

    Each parameter is accessible as an attribute when specified.

    Attributes:
        seeder_mask_size (int): Width in pixels of the region used by the seeder to
            compute the average correlation between a pixel and its neighbors.
        seeder_exclusion_padding (int): Distance for excluding additional pixels
            surrounding segmented cells.
        seeder_grid_size (int): Size of grid bloc per dimension. Seeder maintains only
            the best candidate pixel for each grid block.
        percentage_of_seeds (float[0, 1]): Fraction of candidate seeds to evaluate.
        postprocessor_min_cell_size (int): Lower bound on pixel count of a cell.
        postprocessor_max_cell_size (int): Upper bound on pixel count of a cell.
        postprocessor_preferred_cell_size (int): Pixel count of a typical cell.
        positive_seed_radius (int): Radius of the positive seed square / superpixel.
        negative_seed_circle_radius (int): Radius in pixels of the circle with negative
            seeds.
        negative_seed_circle_count (int): Number of negative seeds.
        gaussian_similarity_alpha (alpha): Decay factor in gaussian similarity function.
        sparse_computation_grid_distance (float): 1 / grid_resolution. Width of each
            block in sparse computation.
        sparse_computation_dimension (int): Dimension of the low-dimensional space in
            sparse computation.
        patch_size (int): Size in pixel of each dimension of the patch.
        _entries (dict): Dict with parameter keys and values. Each parameter value
            (when defined) is also accessible as an attribute.
    """

    def __init__(self, **entries):
        """Initializes HNCcorrConfig object."""

        allowed_parameters = {
            "seeder_mask_size",
            "seeder_exclusion_padding",
            "seeder_grid_size",
            "percentage_of_seeds",
            "postprocessor_min_cell_size",
            "postprocessor_max_cell_size",
            "postprocessor_preferred_cell_size",
            "positive_seed_radius",
            "negative_seed_circle_radius",
            "negative_seed_circle_count",
            "gaussian_similarity_alpha",
            "sparse_computation_grid_distance",
            "sparse_computation_dimension",
            "patch_size",
        }

        for param in entries:
            if param not in allowed_parameters:
                raise ValueError("Parameter %s is not valid." % param)
        self._entries = entries

        for key, value in self._entries.items():
            setattr(self, key, value)

    def __add__(self, other):
        """Combines two configurations and returns a new one.

        If parameters are defined in both configurations, then `other` takes precedence.

        Args:
            other (HNCcorrConfig): Another configuration object.

        Returns:
            HNCcorrConfig: Configuration with combined parameter sets.

        Raises:
            TypeError: When other is not an instance of HNCcorrConfig.
        """
        if not isinstance(other, HNCcorrConfig):
            raise TypeError(
                "other is an instance of %s instead of %s." % (type(other), type(self))
            )

        entries = deepcopy(self._entries)
        entries.update(other._entries)  # pylint: disable=W0212

        return HNCcorrConfig(**entries)


DEFAULT_CONFIG = HNCcorrConfig(
    seeder_mask_size=3,
    seeder_exclusion_padding=4,
    seeder_grid_size=5,
    percentage_of_seeds=0.40,
    postprocessor_min_cell_size=40,
    postprocessor_max_cell_size=200,
    postprocessor_preferred_cell_size=80,
    positive_seed_radius=0,
    negative_seed_circle_radius=10,
    negative_seed_circle_count=10,
    gaussian_similarity_alpha=1.0,
    sparse_computation_grid_distance=1 / 35.0,
    sparse_computation_dimension=3,
    patch_size=31,
)



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
"""Provides example data for HNCcorr."""
import errno
import os
import numpy as np
from tqdm import tqdm
from six.moves import urllib


def load_example_data(  # pylint: disable=C0330
    filedir=".", filename="neurofinder.02.00_agg10.npy", download=True, memmap=False
):
    """Downloads a subsampled copy of the Neurofinder 02.00 dataset.

    Dataset is subsampled in the time dimension, where every 10 frames are replaced by
    a single frame with average intensity values. Resulting data has 800 frames, each
    of 512 x 512 pixels.

    Args:
        filedir (str): Directory for data file. Default is current directory.
        filename (str): Filename for data file.
        download (bool): If True and file does not exist, a new copy is downloaded. If
            False, data array is only loaded.
        memmap (bool): If True, a array-like memory map for the data is returned. If
            False, the data is loaded into memory. Default is False.

    Returns:
        np.array: Returns Numpy array like copy of the movie data.
    """
    url = "https://hnccorr-example-data.s3-us-west-2.amazonaws.com/neurofinder.02.00_agg10.npy"  # pylint: disable=C0301

    if memmap:
        memmap_mode = "r"
    else:
        memmap_mode = None

    filepath = os.path.join(filedir, filename)

    if os.path.isfile(filepath):
        pass
    elif download:
        download_url(url, filedir, filename=filename)
    else:
        raise IOError(
            "The file %s in directory %s does not exist. Set download=True."
            % (filename, os.path.expanduser(filedir))
        )

    return np.load(filepath, mmap_mode=memmap_mode)


def gen_bar_updater():
    """Provides an progress meter. Taken from Pytorch. See licence in source code."""
    # BSD 3-Clause License
    #
    # Copyright (c) Soumith Chintala 2016,
    # All rights reserved.
    #
    # Redistribution and use in source and binary forms, with or without modification,
    # are permitted provided that the following conditions are met:
    #
    # * Redistributions of source code must retain the above copyright notice, this
    # list of conditions and the following disclaimer.
    #
    # * Redistributions in binary form must reproduce the above copyright notice, this
    # list of conditions and the following disclaimer in the documentation and/or other
    # materials provided with the distribution.
    #
    # * Neither the name of the copyright holder nor the names of its contributors may
    # be used to endorse or promote products derived from this software without
    # specific prior written permission.
    #
    # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    # ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    # DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
    # ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    # (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
    # OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    # THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    # NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
    # IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    pbar = tqdm(total=None, unit="MB", mininterval=0.5)

    def bar_update(count, block_size, total_size):
        if pbar.total is None and total_size:
            pbar.total = total_size / (1024.0 ** 2)
        progress_bytes = count * block_size / (1024.0 ** 2)
        pbar.update(progress_bytes - pbar.n)

    return bar_update


def download_url(url, root, filename=None):
    """Download a file from a url and place it in root.

    Taken from Pytorch. See licence in source code.

    Args:
        url (str): URL to download file from
        root (str): Directory to place downloaded file in
        filename (str, optional): Name to save the file under. If None, use the
            basename of the URL
    """
    # BSD 3-Clause License
    #
    # Copyright (c) Soumith Chintala 2016,
    # All rights reserved.
    #
    # Redistribution and use in source and binary forms, with or without modification,
    # are permitted provided that the following conditions are met:
    #
    # * Redistributions of source code must retain the above copyright notice, this
    # list of conditions and the following disclaimer.
    #
    # * Redistributions in binary form must reproduce the above copyright notice, this
    # list of conditions and the following disclaimer in the documentation and/or other
    # materials provided with the distribution.
    #
    # * Neither the name of the copyright holder nor the names of its contributors may
    # be used to endorse or promote products derived from this software without
    # specific prior written permission.
    #
    # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    # ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    # DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
    # ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    # (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
    # OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    # THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    # NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
    # IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    root = os.path.expanduser(root)
    if not filename:
        filename = os.path.basename(url)
    fpath = os.path.join(root, filename)

    # check if path exists.
    try:
        os.makedirs(root)
    except OSError as error:
        if error.errno == errno.EEXIST:
            pass
        else:
            raise

    # downloads file
    try:
        print("Downloading " + url + " to " + fpath)
        urllib.request.urlretrieve(url, fpath, reporthook=gen_bar_updater())
    except urllib.error.URLError as error:
        if url[:5] == "https":
            url = url.replace("https:", "http:")
            print(
                "Failed download. Trying https -> http instead."
                " Downloading " + url + " to " + fpath
            )
            urllib.request.urlretrieve(url, fpath, reporthook=gen_bar_updater())
        else:
            raise error



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
import pytest

from hnccorr.seeds import (
    PositiveSeedSelector,
    NegativeSeedSelector,
    LocalCorrelationSeeder,
)


@pytest.fixture
def LCS():
    return LocalCorrelationSeeder(3, 0.2, 2, 1)


def extract_valid_pixels_10_10(pixels):
    return pixels.intersection({(i, j) for i in range(10) for j in range(10)})


class TestLocalCorrelationSeeder:
    def test_local_corr_seeder(self, LCS, MM):
        LCS.select_seeds(MM)
        assert LCS.next() == (9,)
        assert LCS.next() == (8,)
        assert LCS.next() is None

    def test_local_corr_seeder_grid_size(self, MM):
        lcs = LocalCorrelationSeeder(3, 1.0, 2, 5)
        lcs.select_seeds(MM)
        assert lcs.next() == (9,)
        assert lcs.next() == (4,)
        assert lcs.next() is None

    def test_local_corr_seeder_width_not_divisable_by_grid_size(self, MM):
        lcs = LocalCorrelationSeeder(3, 1.0, 2, 4)
        lcs.select_seeds(MM)
        assert lcs.next() == (9,)
        assert lcs.next() == (7,)
        assert lcs.next() == (3,)
        assert lcs.next() is None

    def test_local_corr_seeder_reset(self, LCS, MM):
        LCS.select_seeds(MM)
        assert LCS.next() == (9,)

        LCS.reset()
        assert LCS.next() == (9,)

    def test_seeder_exclude_pixels(self, LCS, MM):
        LCS.select_seeds(MM)
        assert LCS.next() == (9,)
        LCS.exclude_pixels({(8,)})
        assert LCS.next() is None

    def test_seeder_exclude_pixels_boundary(self, LCS, MM):
        LCS.select_seeds(MM)
        assert LCS.next() == (9,)
        LCS.exclude_pixels({(6,)})
        assert LCS.next() is None

    def test_seeder_reset_excluded_pixels(self, LCS, MM):
        LCS.select_seeds(MM)
        assert LCS.next() == (9,)
        LCS.exclude_pixels({(9,)})
        LCS.reset()
        assert LCS.next() == (9,)

    def test_seeder_select_seeds_should_reset_excluded_pixels(self, LCS, MM):
        LCS.select_seeds(MM)
        LCS.exclude_pixels({(9,)})
        LCS.select_seeds(MM)
        assert LCS.next() == (9,)


class TestPositiveSeedSelector:
    @pytest.mark.parametrize(
        "center_seed, radius, expected_seeds",
        [
            ((0, 0), 0, {(0, 0)}),
            ((0, 0), 1, {(0, 0), (0, 1), (1, 1), (1, 0)}),
            (
                (1, 1),
                1,
                {
                    (0, 0),
                    (0, 1),
                    (0, 2),
                    (1, 0),
                    (1, 1),
                    (1, 2),
                    (2, 0),
                    (2, 1),
                    (2, 2),
                },
            ),
        ],
    )
    def test_positive_seed_selector(
        self, center_seed, radius, expected_seeds, mock_movie
    ):
        mock_movie.num_dimensions = 2
        mock_movie.pixel_size = (10, 10)
        mock_movie.extract_valid_pixels = extract_valid_pixels_10_10

        assert (
            PositiveSeedSelector(radius).select(center_seed, mock_movie)
            == expected_seeds
        )


class TestNegativeSeedSelector:
    @pytest.mark.parametrize(
        "radius, count, center_seed, movie_pixel_size, expected_seeds",
        [
            (0, 8, (5, 5), (10, 10), {(5, 5)}),
            (2, 4, (0, 0), (10, 10), {(0, 2), (2, 0)}),
            (2, 4, (9, 9), (10, 10), {(9, 7), (7, 9)}),
            (
                2,
                8,
                (5, 5),
                (10, 10),
                {(3, 5), (4, 4), (7, 5), (4, 6), (5, 3), (6, 4), (5, 7), (6, 6)},
            ),
            (2, 4, (5, 5), (10, 10), {(3, 5), (7, 5), (5, 3), (5, 7)}),
        ],
    )
    def test_select_negative_seeds(
        self, radius, count, center_seed, movie_pixel_size, expected_seeds, mock_movie
    ):
        mock_movie.num_dimensions = len(movie_pixel_size)
        mock_movie.pixel_size = movie_pixel_size
        mock_movie.extract_valid_pixels = extract_valid_pixels_10_10

        assert (
            NegativeSeedSelector(radius, count).select(center_seed, mock_movie)
            == expected_seeds
        )

    @pytest.mark.parametrize(
        "radius, count, center_seed, movie_pixel_size",
        [(2, 4, (5, 5, 5), (10, 10, 10)), (2, 4, (5,), (10,))],
    )
    def test_select_negative_seeds_invalid_dimension(
        self, radius, count, center_seed, movie_pixel_size, mock_movie
    ):
        mock_movie.num_dimensions = len(movie_pixel_size)
        mock_movie.pixel_size = movie_pixel_size

        with pytest.raises(ValueError):
            NegativeSeedSelector(radius, count).select(center_seed, mock_movie)



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
import pytest

from sparsecomputation import PCA
from hnccorr.movie import Movie, Patch
from hnccorr.base import HNCcorr, Candidate
from hnccorr.graph import (
    GraphConstructor,
    SparseComputationEmbeddingWrapper,
    CorrelationEmbedding,
    exponential_distance_decay,
)
from hnccorr.seeds import (
    PositiveSeedSelector,
    NegativeSeedSelector,
    LocalCorrelationSeeder,
)
from hnccorr.postprocessor import SizePostprocessor
from hnccorr.segmentation import HncParametricWrapper, Segmentation


@pytest.fixture
def matlab_segmentation():
    return Segmentation(
        {
            # First 3 pixels are manually ADDED due to small differences in
            # implementation.
            (285, 433),
            (282, 432),
            (279, 440),
            # Solution MATLAB:
            (280, 433),
            (281, 433),
            (282, 433),
            (283, 433),
            (284, 433),
            (279, 434),
            (280, 434),
            (281, 434),
            (282, 434),
            (283, 434),
            (284, 434),
            (285, 434),
            (286, 434),
            (279, 435),
            (280, 435),
            (281, 435),
            (282, 435),
            (283, 435),
            (284, 435),
            (285, 435),
            (286, 435),
            (287, 435),
            (279, 436),
            (280, 436),
            (281, 436),
            (282, 436),
            (283, 436),
            (284, 436),
            (285, 436),
            (286, 436),
            (287, 436),
            (279, 437),
            (280, 437),
            (281, 437),
            (282, 437),
            (283, 437),
            (284, 437),
            (285, 437),
            (286, 437),
            (287, 437),
            (279, 438),
            (280, 438),
            (281, 438),
            (282, 438),
            (283, 438),
            (284, 438),
            (285, 438),
            (286, 438),
            (287, 438),
            (279, 439),
            (280, 439),
            (281, 439),
            (282, 439),
            (283, 439),
            (284, 439),
            (285, 439),
            (286, 439),
            (280, 440),
            (281, 440),
            (282, 440),
            (283, 440),
            (284, 440),
            (285, 440),
            (286, 440),
            (280, 441),
            (281, 441),
            (282, 441),
            (283, 441),
            (284, 441),
            (285, 441),
            (282, 442),
            (283, 442),
        },
        0.0,
    )


def test_hnccorr_single_segment(mocker, dummy, neurofinder_data, matlab_segmentation):
    seeder = LocalCorrelationSeeder(3, 0.4, 4, 5)
    postprocessor = SizePostprocessor(40, 200, 80)
    segmentor = HncParametricWrapper(0, 100000)
    positive_seed_selector = PositiveSeedSelector(0)

    # negative seed selector is mocked due to a bug in the matlab code.
    # Matlab passes the argument to sin(x)/cos(x) in degrees instead of radians.
    # This results in slightly wrong negative seeds.
    negative_seed_selector = mocker.patch(
        "hnccorr.seeds.NegativeSeedSelector", autospec=True
    )(dummy, dummy)
    negative_seed_selector.select.return_value = {
        (283, 447),
        (273, 435),
        (285, 427),
        (292, 440),
        (278, 445),
        (274, 431),
        (289, 429),
        (289, 444),
        (274, 442),
        (278, 427),
    }

    edge_selector = SparseComputationEmbeddingWrapper(
        3, 1 / 35.0, dimension_reducer=PCA(3)
    )
    graph_constructor = GraphConstructor(
        edge_selector, lambda a, b: exponential_distance_decay(a, b, 1.0)
    )
    patch_size = 31

    H = HNCcorr(
        seeder,
        postprocessor,
        segmentor,
        positive_seed_selector,
        negative_seed_selector,
        graph_constructor,
        Candidate,
        Patch,
        CorrelationEmbedding,
        patch_size,
    )

    H.movie = Movie("Neurofinder02.00", neurofinder_data)

    center_seed = (282, 436)
    c = Candidate(center_seed, H)
    best_segmentation = c.segment()

    assert best_segmentation.selection == matlab_segmentation.selection
    assert best_segmentation.weight == pytest.approx(matlab_segmentation.weight)



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
import pytest
import numpy as np
from hnccorr.graph import (
    CorrelationEmbedding,
    exponential_distance_decay,
    GraphConstructor,
    SparseComputationEmbeddingWrapper,
)


@pytest.fixture
def mock_patch(mocker, dummy):
    return mocker.patch("hnccorr.movie.Patch", autospec=True)(dummy, dummy, dummy)


@pytest.fixture
def mock_edge_selector(mocker, dummy):
    return mocker.patch(
        "hnccorr.graph.SparseComputationEmbeddingWrapper", autospec=True
    )(dummy, dummy)


@pytest.fixture
def mock_embedding(mocker, dummy):
    return mocker.patch("hnccorr.graph.CorrelationEmbedding", autospec=True)(dummy)


@pytest.fixture
def CE1(mock_patch):
    mock_patch.pixel_shape = (7,)
    mock_patch.__getitem__.return_value = np.array(
        [
            [1, 1, 1, 1, 1, 1, 1],
            [-1, -1, -1, -1, -1, -1, -1],
            [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
        ]
    )

    return CorrelationEmbedding(mock_patch)


@pytest.fixture
def CE2(mock_patch):
    mock_patch.pixel_shape = (3, 3)
    mock_patch.__getitem__.return_value = np.zeros((3, 3))
    return CorrelationEmbedding(mock_patch)


class TestEmbedding:
    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    def test_embedding_embedding(self, CE1, CE2, mock_patch):

        np.testing.assert_allclose(
            CE1.embedding[0],
            [
                1.0,
                0.99833749,
                0.99339927,
                0.98532928,
                0.9743547,
                0.96076892,
                0.94491118,
            ],
        )

        np.testing.assert_allclose(CE2.embedding[(0, 0, slice(None, None))], [0, 0, 0])

    def test_embedding_get_vector(self, CE1):
        CE1.embedding = np.array([[0.0, 1.0], [-2.0, 0.0]])
        np.testing.assert_allclose(CE1.get_vector((0,)), np.array([0.0, -2.0]))


def test_exponential_distance_decay():
    alpha = 0.5

    exponential_distance_decay(
        np.array([0.0, -2.0]), np.array([[1.0, 0.0]]), alpha
    ) == pytest.approx(np.exp(-0.25))


class TestGraphConstructor:
    def test_graph_constructor(self, mock_patch, mock_edge_selector, mock_embedding):
        mock_edge_selector.select_edges.return_value = [((0,), (1,)), ((0,), (2,))]
        mock_patch.enumerate_pixels.return_value = [(i,) for i in range(7)]
        mock_patch.to_movie_coordinate = lambda x: x
        mock_embedding.get_vector = lambda x: x

        GC = GraphConstructor(mock_edge_selector, lambda a, b: b[0])
        G = GC.construct(mock_patch, mock_embedding)

        num_nodes = 7
        assert len(G.nodes) == num_nodes
        for i in range(num_nodes):
            assert (i,) in G.nodes

        assert len(G.edges) == 4
        assert ((0,), (1,)) in G.edges
        assert ((0,), (2,)) in G.edges
        assert ((2,), (0,)) in G.edges
        assert ((1,), (0,)) in G.edges
        assert G[(0,)][(1,)]["weight"] == 1
        assert G[(0,)][(1,)]["weight"] == 1
        assert G[(2,)][(0,)]["weight"] == 2
        assert G[(2,)][(0,)]["weight"] == 2

    def test_graph_constructor_nodes_offset_from_zero(
        self, mock_patch, mock_edge_selector, mock_embedding
    ):
        all_pixels = {(2,), (3,), (4,), (5,), (6,), (7,), (8,)}
        mock_patch.enumerate_pixels.return_value = all_pixels
        mock_patch.to_movie_coordinate = lambda x: (2,) if x == (0,) else (3,)
        mock_embedding.get_vector = lambda x: x

        mock_edge_selector.select_edges.return_value = [((0,), (1,))]

        GC = GraphConstructor(mock_edge_selector, lambda x, y: 1)
        graph = GC.construct(mock_patch, mock_embedding)

        assert set(graph.nodes) == all_pixels
        assert graph[(2,)][(3,)]["weight"] == 1
        assert graph[(3,)][(2,)]["weight"] == 1


class TestSparseComputationEmbeddingWrapper:
    def test_sparse_computation_select_edges(self, mock_embedding):
        mock_embedding.embedding = np.array(
            [[-1, 0], [-0.9, 0], [0, 0], [0.9, 0], [1, 0]]
        ).T
        assert SparseComputationEmbeddingWrapper(2, 0.2).select_edges(
            mock_embedding
        ) == {((0,), (1,)), ((3,), (4,))}

    def test_sparse_computation_with_dimension_reducer(self, mock_embedding):
        mock_embedding.embedding = np.array(
            [[-1, 0], [-0.9, 0], [0, 0], [0.9, 0], [1, 0]]
        ).T

        class MockDimReducer:
            def fit_transform(self, data, **kwargs):
                return data

        assert SparseComputationEmbeddingWrapper(
            2, 0.2, dimension_reducer=MockDimReducer()
        ).select_edges(mock_embedding) == {((0,), (1,)), ((3,), (4,))}



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
"""HNCcorr components related to the similarity graph."""

import networkx as nx
import numpy as np
from sparsecomputation import SparseComputation as SC
from sparsecomputation import ApproximatePCA

from hnccorr.utils import add_time_index


class CorrelationEmbedding:
    """Computes correlation feature vector for each pixel.

    Embedding provides a representation of a pixel in terms of feature vector. The
    feature vector for the CorrelationEmbedding is a vector of pairwise correlations to
    each (or some) pixel in the patch.

    If the correlation is not defined due to a pixel with zero variance, then the
    corelation is set to zero.


    Attributes:
        embedding (np.array): (D, N_1, N_2, ..) array of pairwise correlations, where D
            is the dimension of the embedding and N_1, N_2, .. are the pixel shape of
            the patch.
    """

    def __init__(self, patch):
        """Initializes a CorrelationEmbedding object.

        See class description for details.

        Args:
            patch (Patch): Subregion of movie for which the correlation embedding is
                computed.
        """
        data = patch[:].reshape(-1, np.product(patch.pixel_shape))
        self.embedding = np.corrcoef(data.T).reshape(-1, *patch.pixel_shape)
        self.embedding[np.isnan(self.embedding)] = 0

    def get_vector(self, pixel):
        """Retrieve feature vector of pixel.

        Args:
            pixel (tuple): Coordinate of pixel.

        Returns:
            np.array: Feature vector of pixel.
        """
        return self.embedding[add_time_index(pixel)]


def exponential_distance_decay(feature_vec1, feature_vec2, alpha):
    """Computes ``exp(- alpha / n || x_1 - x_2 ||^2_2)`` for x_1, x_2 in R^n."""
    num_frames = float(feature_vec1.shape[0])
    return np.exp(
        -alpha * np.linalg.norm(feature_vec1 - feature_vec2) ** 2 / num_frames
    )


class GraphConstructor:
    """Graph constructor over a set of pixels.

    Constructs a similarity graph over the set of pixels in a patch. Edges are selected
    by an edge_selector and the similarity weight associated with each edge is computed
    with the weight_function. Edge weights are stored under the attribute ``weight``.

    A directed graph is used for efficiency. That is, arcs (i,j) and (j,i) are used to represent edge [i,j].

    Attributes:
        _edge_selector (EdgeSelector): Object that constructs the edge set of the graph.
        _weight_function (function): Function that computes the edge weight between two
            pixels. The function should take as input two 1-dimensional numpy arrays,
            representing the feature vectors of the two pixels. The function should
            return a float between 0 and 1.
    """

    def __init__(self, edge_selector, weight_function):
        """Initializes a graph constructor."""
        self._edge_selector = edge_selector
        self._weight_function = weight_function

    def construct(self, patch, embedding):
        """Constructs similarity graph for a given patch.

        See class description.

        Args:
            patch (Patch): Defines subregion and pixel set for the graph.
            embedding (CorrelationEmbedding): Provides feature vectors associated with
                each pixel in the patch.

        Returns:
            nx.DiGraph: Similarity graph over pixels in patch.
        """
        graph = nx.DiGraph()

        graph.add_nodes_from(patch.enumerate_pixels())

        for node1, node2 in self._edge_selector.select_edges(embedding):
            node1_movie = patch.to_movie_coordinate(node1)
            node2_movie = patch.to_movie_coordinate(node2)
            weight = self._weight_function(
                embedding.get_vector(node1), embedding.get_vector(node2)
            )
            # add arc in both directions
            graph.add_edge(node1_movie, node2_movie, weight=weight)
            graph.add_edge(node2_movie, node1_movie, weight=weight)

        return graph


class SparseComputationEmbeddingWrapper:
    """Wrapper for SparseComputation that accepts an embedding.

    Attributes:
        _sc (SparseComputation): SparseComputation object.
    """

    def __init__(self, dim_low, distance, dimension_reducer=None):
        """Initializes a SparseComputationEmbeddingWrapper instance.

        Args:
            dim_low (int): Dimension of the low-dimensional space in sparse computation.
            distance (float): 1 / grid_resolution. Defines the size of the grid blocks
                in sparse computation.
            dimension_reducer (DimReducer): Provides dimension reduction for sparse
                computation. By default, approximate principle component analysis is
                used.

        Returns:
            SparseComputationEmbeddingWrapper

        """
        if dimension_reducer is None:
            dimension_reducer = ApproximatePCA(int(dim_low))

        self._sc = SC(dimension_reducer, distance=distance)

    def select_edges(self, embedding):
        """Selects relevant pairwise similarities with sparse computation.

        Determines the set of relevant pairwise similarities based on the sparse
        computation algorithm. See sparse computation for details. Pixel coordinates
        are with respect to the index of the embedding.

        Args:
            embedding (CorrelationEmbedding): Embedding of pixels into feature vectors.

        Returns:
            list(tuple): List of relevant pixel pairs.
        """
        shape = embedding.embedding.shape[1:]
        data = embedding.embedding.reshape(-1, np.product(shape)).T

        pairs = self._sc.select_pairs(data)

        return set(
            (np.unravel_index(a, shape), np.unravel_index(b, shape)) for a, b in pairs
        )



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
import pytest
import os
import numpy as np

from hnccorr.movie import Patch
from hnccorr.segmentation import Segmentation
from hnccorr.base import Candidate

TEST_DATA_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "../test_data"
)


@pytest.fixture
def dummy():
    return "Dummy"


@pytest.fixture
def mock_movie(mocker, dummy):
    movie = mocker.patch("hnccorr.movie.Movie", autospec=True)(dummy, dummy)
    return movie


@pytest.fixture
def MM(mock_movie):
    mock_movie.num_frames = 3
    mock_movie.pixel_shape = (10,)
    mock_movie.num_dimensions = 1

    def data(self, key):
        A = np.zeros((3, 10))
        A[0, :] = 1
        A[1, :] = -1
        A[2, :] = np.arange(10) / 10
        return A.__getitem__(key)

    mock_movie.__getitem__ = data
    mock_movie.is_valid_pixel_coordinate = lambda x: x[0] >= 0 and x[0] < 10

    return mock_movie


@pytest.fixture
def MM2():
    class MockMovie:
        def __init__(self):
            self.num_dimensions = 2
            self.num_frames = 1

            self._A = np.zeros((2, 10, 10))
            self.pixel_shape = self._A.shape[1:]

        def __getitem__(self, key):

            return self._A.__getitem__(key)

        def is_valid_pixel_index(self, index):
            return index[0] >= 0 and index[0] < 10 and index[1] >= 0 and index[1] < 10

    return MockMovie()


@pytest.fixture
def pos_seeds():
    return {(4,), (5,), (6,)}


@pytest.fixture
def simple_segmentation():
    return Segmentation({(0, 1)}, 1.0)


@pytest.fixture
def segmentor_simple_segmentation(simple_segmentation):
    class MockSegmentor:
        def solve(self, graph, pos_seeds, neg_seeds):
            return [simple_segmentation]

    return MockSegmentor()


@pytest.fixture
def simple_positive_seed_selector():
    class MockPositiveSeedSelector:
        def select(center_seed):
            return center_seed

    return MockPositiveSeedSelector()


@pytest.fixture
def postprocessor_select_first():
    class MockPostProcessor:
        def select(self, segmentations):
            return segmentations[0]

    return MockPostProcessor()


@pytest.fixture
def seeder_fixed_val():
    class MockSeeder:
        def __init__(self):
            self.called = False
            self.return_val = 1

        def select_seeds(self, movie):
            pass

        def exclude_pixels(self, pixel_set):
            pass

        def next(self):
            if self.called:
                return None
            else:
                self.called = True
                return self.return_val

        def reset(self):
            self.called = False

    return MockSeeder()



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
from conftest import TEST_DATA_DIR
import numpy as np
import matplotlib.pyplot as plt

from hnccorr import HNCcorr, Movie, HNCcorrConfig
from hnccorr.example import load_example_data


def test_example():
    movie = Movie(
        "Example movie",
        load_example_data(
            filedir=TEST_DATA_DIR
        ),  # downloads sample Neurofinder dataset.
    )  # See documentation for alternatives
    H = HNCcorr.from_config(
        HNCcorrConfig(percentage_of_seeds=0.025)
    )  # Initialize HNCcorr with default configuration
    H.segment(movie)  # Identify cells in movie

    H.segmentations  # List of identified cells
    H.segmentations_to_list()  # Export list of cells (for Neurofinder)]

    A = np.zeros(movie.pixel_shape)
    for segmentation in H.segmentations:
        for i, j in segmentation.selection:
            A[i, j] += 1

    plt.figure(figsize=(6, 6))
    plt.imshow(A)
    plt.savefig("test cells neurofinder02.00.png")

    assert len(H.segmentations) >= 10



from conftest import TEST_DATA_DIR

from hnccorr.example import load_example_data


def test_example_data_loader():
    data = load_example_data(filedir=TEST_DATA_DIR, memmap=True)
    assert data.shape == (800, 512, 512)



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
import os

from conftest import TEST_DATA_DIR

from hnccorr.utils import (
    four_neighborhood,
    generate_pixels,
    add_offset_to_coordinate,
    add_offset_set_coordinates,
    add_time_index,
    list_images,
    eight_neighborhood,
)


def test_add_offset():
    assert add_offset_to_coordinate((1, 2), (3, 4)) == (4, 6)


def test_add_set_offset():
    assert add_offset_set_coordinates({(0, 1), (1, 1)}, (2, 2)) == {(2, 3), (3, 3)}


def test_add_time_index():
    assert add_time_index((5, 4)) == (slice(None, None), 5, 4)


def test_list_images():
    images = list_images(TEST_DATA_DIR)
    expected_images = map(
        lambda x: os.path.join("./test_data/simple_movie", x),
        ["simple_movie00000.tif", "simple_movie00001.tif", "simple_movie00002.tif"],
    )

    for i, e in zip(images, expected_images):
        assert os.path.abspath(i) == os.path.abspath(e)


def test_eight_neighborhood():
    assert eight_neighborhood(1, 1) == {(-1,), (0,), (1,)}
    assert eight_neighborhood(1, 2) == {(-2,), (-1,), (0,), (1,), (2,)}
    assert eight_neighborhood(2, 1) == {
        (-1, -1),
        (-1, 0),
        (-1, 1),
        (0, -1),
        (0, 0),
        (0, 1),
        (1, -1),
        (1, 0),
        (1, 1),
    }


def test_four_neighborhood():
    assert four_neighborhood(1) == {(-1,), (0,), (1,)}
    assert four_neighborhood(2) == {(-1, 0), (0, 0), (1, 0), (0, -1), (0, 1)}


def test_generate_pixles():
    assert set(generate_pixels((1, 2))) == {(0, 0), (0, 1)}



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
"""Helper functions for HNCcorr."""

import glob
import os
from itertools import product


def add_offset_set_coordinates(iterable, offset):
    """Adds a fixed offset to all pixel coordinates in a set.

    Args:
        coordinates (set): Set of pixel coordinates. Each pixel coordinate is a tuple.
        offset (tuple): Offset to add to each pixel coordinate. Tuple should be of the
            same length as the tuples in `coordinates`.

    Returns:
        set: Set of updated coordinates.

    Example:
        .. code-block:: python

            >>> add_offset_set_coordinates({(5, 2), (4, 7)}, (2, 2))
            {(7, 4), (6, 9)}
    """
    return set(add_offset_to_coordinate(c, offset) for c in iterable)


def add_offset_to_coordinate(coordinate, offset):
    """Offsets pixel coordinate by another coordinate.

    Args:
        coordinate (tuple): Pixel coordinate to offset.
        offset (tuple): Offset to add to coordinate. Must be of the same length.

    Example:
        .. code-block:: python

            >>> add_offset_to_coordinate((5, 3, 4), (1, -1, 3))
            (6, 2, 7)
    """
    return tuple(a + b for a, b in zip(coordinate, offset))


def add_time_index(index):
    """Inserts a full slice as the first dimension of an index for e.g. numpy.

    Args:
        index (tuple): Index for e.g. numpy array.

    Returns:
        tuple: New index with additional dimension.

    Example:
        .. code-block:: python

            >>> add_time_index((5, :3))
            (:, 5, :3)
    """
    return (slice(None, None),) + index


def list_images(folder):
    """Lists and sorts tiff images in a folder.

    Images are sorted in ascending order based on filename.

    Caution:
        Filenames are sorted as strings. Note that ``200.tiff`` is sorted before
        ``5.tiff``. Pad image filenames with zeros to prevent this: ``005.tiff``.

    Args:
        folder: folder containing tiff image files.

    Returns:
        list: Sorted list of paths of tiff files in folder.
    """
    files_tif = glob.glob(os.path.join(folder, "*.tiff"))
    return sorted(files_tif)


def four_neighborhood(num_dims):
    """Returns all neighboring pixels of zero that differ in at most one coordinate.

    Includes zero coordinate itself.

    Args:
        num_dims (int): Number of dimensions for the coordinates.

    Returns:
        set: Set of pixel coordinates.

    Example:
        .. code-block:: python

            >>> four_neighborhood(1)
            [(-1,), (0,), (1,)]
            >>> eight_neighborhood(2)
            [(-1, 0), (0, -1), (0, 0), (0, 1), (1, 0)]

    """
    neighbors = []
    for dim, change in product(range(num_dims), (-1, 0, 1)):
        shift = [0] * num_dims
        shift[dim] = change
        neighbors.append(tuple(shift))
    return set(neighbors)


def eight_neighborhood(num_dims, max_radius):
    """Returns all coordinates within a given L-infinity distance of zero.

    Includes zero coordinate itself.

    Args:
        num_dims (int): Number of dimensions for the coordinates.
        max_radius (int): Largest L-infinity distance allowed.

    Returns:
        set: Set of pixel coordinates.

    Example:
        .. code-block:: python

            >>> eight_neighborhood(1, 1)
            [(-1,), (0,), (1,)]
            >>> eight_neighborhood(2, 1)
            [
                (-1, -1), (-1, 0), (-1, 1), (0, -1), (0, 0),
                (0, 1), (1, -1), (1, 0), (1, 1)
            ]

    """
    return set(product(range(-max_radius, max_radius + 1), repeat=num_dims))


def generate_pixels(shape):
    """Enumerate all pixel coordinates for a movie/patch.

    Args:
        shape (tuple): Shape of movie. Number of pixels in each dimension.

    Returns:
        Iterator: Iterates over all pixels.

    Example:
        .. code-block:: python

            >>> generate_pixels((2,2))
            [(0, 0), (0, 1), (1, 0), (1, 1)]
    """
    return product(*[range(n) for n in shape])



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
"""HNC and segmentation related components in HNCcorr."""

from copy import deepcopy
from itertools import product
import networkx as nx
import numpy as np
from scipy.ndimage.morphology import binary_fill_holes

from closure.hnc import HNC as HNC_Closure

from hnccorr.utils import four_neighborhood


class HncParametricWrapper:
    r"""Wrapper for solving the Hochbaum Normalized Cut (HNC) problem on a graph.

    Given an undirected graph :math:`G = (V, E)` with edge weights :math:`w_{ij} \ge 0`
    for :math:`[i,j] \in E`, the linearized HNC problem is defined as:

    .. math::
        \min_{\emptyset \subset S \subset V} \sum_{\substack{[i,j] \in E,\\ i \in S,\\
        j \in V \setminus S}} w_{ij} - \lambda \sum_{i \in S} d_i,

    where $d_i$ the degree of node :math:`i \in V` and :math:`\lambda \ge 0` provides
    the trade-off between the two objective terms.

    See closure package for solution method.
    """

    def __init__(self, lower_bound, upper_bound):
        """Initializes HncParametricWrapper object."""
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound

    @staticmethod
    def _construct_segmentations(source_sets, breakpoints):
        """Constructs a list of segmentations from output HNC.

        Each source set and corresponding lambda upper bound is replaced with a
        Segmentation object where the selection matches the source set and the weight
        parameter matches the upper bound of the lambda range.

        Args:
            source_sets (list[set]): List of source sets for each lambda range.
            breakpoints (list[float]): List of upper bounds on the lambda range for
                which the corresponding source set is optimal.

        Returns:
            list[Segmentation]: List of segmentations.

        """
        return [
            Segmentation(selection, weight)
            for selection, weight in zip(source_sets, breakpoints)
        ]

    def solve(self, graph, pos_seeds, neg_seeds):
        """Solves an instance of the HNC problem for all values of lambda.

        Solves the HNC clustering problem on `graph` for all values of lambda
        simultaneously. See class description for a definition of HNC.

        Args:
            graph (nx.Graph): Directed similarity graph with non-negative edge
                weights. Edge [i,j] is represented by two directed arcs (i,j) and
                (j,i). Edge weights must be defined via the attribute `weight`.
            pos_seeds (set): Set of nodes in graph that must be part of the cluster.
            neg_seeds (set): Set of nodes in graph that must be part of the complement.

        Returns:
            list[Segmentation]: List of optimal clusters for each lambda range.

        Caution:
            Class modifies graph for performance. Pass a copy to prevent any issues.
        """
        hnc = HNC_Closure(graph, pos_seeds, neg_seeds, arc_weight="weight")
        source_sets, breakpoints = hnc.solve_parametric(
            self._lower_bound, self._upper_bound
        )
        return self._construct_segmentations(source_sets, breakpoints)


class Segmentation:
    """A set of pixels identified by HNC as a potential cell footprint.

    Attributes:
        selection (set): Pixels in the spatial footprint. Each pixel is represented as
            a tuple.
        weight (float): Upper bound on the lambda coefficient for which this
            segmentation is optimal.
    """

    def __init__(self, selection, weight):
        """Initializes a Segmentation object."""
        self.selection = set(selection)
        self.weight = weight

    def __eq__(self, other):
        """Compares two Segmentation objects."""
        if isinstance(other, Segmentation):
            return (self.selection == other.selection) and (self.weight == other.weight)

        return False

    def clean(self, positive_seeds, movie_pixel_shape):
        """Cleans Segmentation by selecting a connected component and filling holes.

        The Segmentation is decomposed into connected components by considering
        horizontal or vertical adjacent pixels as neighbors. The connected component
        with the most positive seeds is selected. Any holes in the selected component
        are added to the selection.

        Args:
            positive_seeds (set): Pixels that are contained in the
                spatial footprint. Each pixel is represented by a tuple.
            movie_pixel_shape (tuple): Pixel resolution of the movie.

        Returns:
            Segmentation: A new Segmentation object with the same weight.
        """
        improved_segmentation = self.select_max_seed_component(positive_seeds)
        return improved_segmentation.fill_holes(movie_pixel_shape)

    def select_max_seed_component(self, positive_seeds):
        """Selects the connected component of selection that contains the most seeds.

        The Segmentation is decomposed into connected components by considering
        horizontal or vertical adjacent pixels as neighbors. The connected component
        with the most positive seeds is selected.

        Args:
            positive_seeds (set): Pixels that are contained in the
                spatial footprint. Each pixel is represented by a tuple.

        Returns:
            Segmentation: A new Segmentation object with the same weight.
        """

        # get an arbitrary element from seeds to compute dimension
        num_dims = len(next(iter(self.selection)))
        neighbors = four_neighborhood(num_dims)

        graph = nx.Graph()
        graph.add_nodes_from(self.selection)

        for index, shift in product(self.selection, neighbors):
            neighbor = tuple(map(lambda a, b: a + b, index, shift))
            if neighbor in graph:
                graph.add_edge(index, neighbor)

        components = list(nx.connected_components(graph))

        overlap = [len(c.intersection(positive_seeds)) for c in components]

        best_component = components[np.argmax(overlap)]

        return Segmentation(best_component, self.weight)

    def fill_holes(self, movie_pixel_shape):
        """Fills holes in the selection.

        Args:
            movie_pixel_shape (tuple): Pixel resolution of the movie.

        Returns:
            Segmentation: A new Segmentation object with the same weight.
        """
        mask = np.full(movie_pixel_shape, False, dtype=np.bool)

        indices = tuple(zip(*self.selection))
        mask[indices] = True

        filled_mask = binary_fill_holes(mask)

        index_arrays = [a.tolist() for a in np.where(filled_mask)]
        return Segmentation(set(zip(*index_arrays)), self.weight)



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
"""Seed related components of HNCcorr."""

from math import sin, cos, pi
import itertools
import numpy as np

from hnccorr.utils import (
    add_offset_set_coordinates,
    add_time_index,
    eight_neighborhood,
    generate_pixels,
)


class LocalCorrelationSeeder:
    """Provide seeds based on the correlation of pixels to their local neighborhood.

    Seed pixels are selected based on the average correlation of the pixel to its local
    neighborhood.For each block of `grid_size` by `grid_size` pixels, the pixel with
    the highest average local correlation is selected. The remaining pixels in each
    block are discarded. From the remaining pixels, a fraction of `_seed_fraction`
    pixels, those with the highest average local correlation, are kept and attempted
    for segmentation.

    The local neighborhood of each pixel consist of the pixels in a square of width
    `_neighborhood_size` centered on the pixels. Pixel coordinates outside the boundary
    of the movie are ignored.


    Attributes:
        _current_index (int): Index of next seed in `_seeds` to return.
        _excluded_pixels (set): Set of pixel coordinates to excluded as future seeds.
        _grid_size (int): Number of pixels per dimension in a block.
        _keep_fraction (float): Percentage of candidate seed pixels to attempt for
            segmentation. All other candidate seed pixels are discarded.
        _movie (Movie): Movie to segment.
        _neighborhood_size (int): Width in pixels of the local neighborhood of a pixel.
        _padding (int): L-infinity distance for determining which pixels should be
            padded to the exclusion set in `exclude_pixels()`.
        _seeds (list[tuple]): List of candidate seed coordinates to return.
    """

    def __init__(self, neighborhood_size, keep_fraction, padding, grid_size):
        """Initializes a LocalCorrelationSeeder object."""
        self._current_index = None
        self._excluded_pixels = set()
        self._keep_fraction = keep_fraction
        self._movie = None
        self._neighborhood_size = neighborhood_size
        self._padding = padding
        self._seeds = None
        self._grid_size = grid_size

    def select_seeds(self, movie):
        """Identifies candidate seeds in movie.

        Initializes list of candidate seeds in the movie. See class description for
        details. Seeds can be accessed via the :meth:`~.LocalCorrelationSeeder.next`
        method.

        Args:
            movie (Movie): Movie object to segment.

        Returns:
            None
        """
        self._movie = movie

        num_dimensions = self._movie.num_dimensions
        # helpful constants
        max_shift = int((self._neighborhood_size - 1) / 2)

        # generate all offsets of neighbors
        neighbor_offsets = eight_neighborhood(num_dimensions, max_shift)
        # remove point as neighbor
        neighbor_offsets = neighbor_offsets - {(0,) * num_dimensions}

        mean_neighbor_corr = {}

        for pixel in generate_pixels(self._movie.pixel_shape):
            # compute neighbors
            neighbors = add_offset_set_coordinates(neighbor_offsets, pixel)

            # extract data for valid neighbors
            valid_neighbors = [
                neighbor
                for neighbor in neighbors
                if self._movie.is_valid_pixel_coordinate(neighbor)
            ]

            # store average correlation
            mean_neighbor_corr[pixel] = self._compute_average_local_correlation(
                pixel, valid_neighbors
            )

        best_per_grid_block = self._select_best_per_grid_block(mean_neighbor_corr)

        best_per_grid_block_sorted = sorted(
            [(key, val) for key, val in best_per_grid_block.items()],
            key=lambda x: x[1],
            reverse=True,
        )

        num_keep = int(self._keep_fraction * len(best_per_grid_block_sorted))

        # store best seeds
        self._seeds = [seed for seed, _ in best_per_grid_block_sorted[:num_keep]]
        self.reset()

    def _compute_average_local_correlation(self, pixel, valid_neighbors):
        """Compute average correlation between pixel and neighbors."""
        pixel_data = self._movie[add_time_index(pixel)].reshape(1, -1)

        # extract data for valid neighbors
        neighbors_data = []
        for neighbor in valid_neighbors:
            neighbors_data.append(self._movie[add_time_index(neighbor)].reshape(1, -1))
        neighbors_data = np.concatenate(neighbors_data, axis=0)

        # compute correlation to each neighbor (corrcoef concatenates the
        # two vectors so we extract last row except for last element)
        neighbors_corr = np.corrcoef(neighbors_data, pixel_data)[-1, :-1]

        # store average correlation
        return np.mean(neighbors_corr)

    def _select_best_per_grid_block(self, scores):
        """Selects pixel with highest score in a block of grid_size pixels per dim.

        """
        pixel_shape = self._movie.pixel_shape
        num_dimensions = self._movie.num_dimensions

        best_pixels = {}

        # identify all pixels in a block relatve to top left pixel as base coordinate
        shifts = set(
            itertools.product(*(range(self._grid_size) for _ in range(num_dimensions)))
        )

        # loop over all base coordinates of the grid, e.g. (0, 0), (5, 0), (0, 5),
        # (5, 5), (10, 0), ... for grid_size = 5
        for base_coordinate in itertools.product(
            *(range(0, pixel_shape[i], self._grid_size) for i in range(num_dimensions))
        ):
            # list of all coordinates in the block
            coordinates_in_block = add_offset_set_coordinates(shifts, base_coordinate)
            # select coordinate in block with highest score
            best_coordinate, best_value = max(
                [
                    (coordinate, scores[coordinate])
                    for coordinate in coordinates_in_block
                    if coordinate in scores
                ],
                key=lambda x: x[1],
            )

            best_pixels[best_coordinate] = best_value

        return best_pixels

    def exclude_pixels(self, pixels):
        """Excludes pixels from being returned by `next()` method.

        All pixels within in the set `pixels` as well as pixels that are within an L-
        infinity distance of `_padding` from any excluded pixel are excluded as seeds.

        Method enables exclusion of pixels in previously segmented cells from serving
        as new seeds. This may help to prevent repeated segmentation of the cell.

        Args:
            pixels (set): Set of pixel coordinates to exclude.

        Returns:
            None
        """
        neighborhood = eight_neighborhood(self._movie.num_dimensions, self._padding)

        padded_pixel_sets = [
            add_offset_set_coordinates(neighborhood, pixel) for pixel in pixels
        ]

        self._excluded_pixels = self._excluded_pixels.union(
            pixels.union(*padded_pixel_sets)
        )

    def next(self):
        """Provides next seed pixel for segmentation.

        Returns the movie coordinates of the next available seed pixel for
        segmentation. Seed pixels that have previously been excluded will be ignored.
        Returns None when all seeds are exhausted.

        Returns:
            tuple or None: Coordinates of next seed pixel. None if no seeds remaining.
        """
        while self._current_index < len(self._seeds):
            center_seed = self._seeds[self._current_index]
            self._current_index += 1

            if center_seed not in self._excluded_pixels:
                return center_seed

        return None

    def reset(self):
        """Reinitialize the sequence of seed pixels and empties `_excluded_seeds`."""
        self._current_index = 0
        self._excluded_pixels = set()


class NegativeSeedSelector:
    """Selects negative seed pixels uniformly from a circle around center seed pixel.

    Selects `_count` pixels from a circle centered on the center seed pixel with radius
    `_radius`. The selected pixels are spread uniformly over the circle. Non-integer
    pixel indices are rounded to the closest (integer) pixel. Currently only
    2-dimensional movies are supported.

    Attributes:
        _radius (float): L2 distance to center seed.
        _count (int): Number of negative seed pixels to select.
    """

    def __init__(self, radius, count):
        self._radius = radius
        self._count = count

    def select(self, center_seed, movie):
        """Selects negative seed pixels.

        Args:
            center_seed (tuple): Center seed pixels.
            movie (Movie): Movie for segmentation.

        Returns:
            set: Set of negative seed pixels. Each pixel is denoted by a tuple.
        """
        if movie.num_dimensions != 2:
            raise ValueError("Only 2-dimensional movies are currently supported.")

        # Determine uniform locations of pixels on the circle in radians.
        angle_step = 2 * pi / float(self._count)
        angles = [i * angle_step for i in range(self._count)]
        offsets = {
            (round(self._radius * cos(x)), round(self._radius * sin(x))) for x in angles
        }
        negative_seeds = add_offset_set_coordinates(offsets, center_seed)

        # check if seeds are within boundaries
        return movie.extract_valid_pixels(negative_seeds)


class PositiveSeedSelector:
    """Selects positive seed pixels in a square centered on `center_seed`.

    Selects all pixels in a square centered on `center_seed` as positive seeds. A pixel
    is selected if it is within a Chebyshev distance (L-Inf) of `_max_distance` from the
    center seed pixel.

    Attributes:
        _max_distance (int): Maximum L-Inf distance allowed.
    """

    def __init__(self, max_distance):
        self._max_distance = max_distance

    def select(self, center_seed, movie):
        """Selects positive seeds.

        Args:
            center_seed (tuple): Center seed pixel.
            movie (Movie): Movie for segmentation.

        Returns:
            set: Set of positive seed pixels. Each pixel is denoted by a tuple.
        """
        offsets = eight_neighborhood(movie.num_dimensions, self._max_distance)
        # compute positive seeds
        positive_seeds = add_offset_set_coordinates(offsets, center_seed)
        # check if seeds are within boundaries
        return movie.extract_valid_pixels(positive_seeds)
